@@ -1,3 +1,5 @@
+/*#define USE_CURSES*/
+#define BYTE_SWAP_DEBUG
 /*
 
 			Speak Freely for Unix
@@ -8,6 +10,7 @@
 */
 
 #include "speakfree.h"
+#include "version.h"
 
 /*  Destination host descriptor.  */
 
@@ -22,6 +25,10 @@ struct destination {
     char vatdeskey[9];		      /* Destination VAT DES key, if any */
     char ideakey[17];		      /* Destination IDEA key, if any */
     char pgpkey[17];		      /* Destination PGP key, if any */
+    char blowfish_spec; 	      /* Nonzero if Blowfish key specified */
+#ifdef CRYPTO
+    BF_KEY blowfishkey; 	      /* Destination Blowfish key, if any */
+#endif
     char *otp;			      /* Key file */
 };
 
@@ -30,6 +37,12 @@ static int sock;		      /* Communication socket */
 static struct destination *dests = NULL, *dtail;
 static int compressing = FALSE;       /* Compress sound: simple 2X */
 static int gsmcompress = TRUE;	      /* GSM compress buffers */
+#ifdef OBSOLETE
+static int gsm_byte_order_kludge = TRUE; /* Transition to correct byte order for GSM ? */
+#endif
+#ifdef BYTE_SWAP_DEBUG
+static int gsm_byte_order_debug = FALSE; /* Send GSM length field with swapped bytes */
+#endif
 static int lpccompress = FALSE;       /* LPC compress buffers */
 static int lpc10compress = FALSE;     /* LPC-10 compress buffers */
 static int adpcmcompress = FALSE;     /* ADPCM compress buffers */
@@ -53,6 +66,7 @@ static int agc = FALSE; 	      /* Automatic gain control active ? */
 static int rgain = 33;		      /* Current recording gain level */
 static int spurt = TRUE;	      /* True for first packet of talk spurt */
 static int debugging = FALSE;	      /* Debugging enabled here and there ? */
+static FILE *audioDumpFile = NULL;    /* Audio dump file, if any */
 static int havesound = FALSE;         /* True if we've acquired sound input */
 static char hostname[20];	      /* Host name to send with packets */
 static int loopback = FALSE;	      /* Remote loopback mode */
@@ -75,10 +89,12 @@ static char currtpkey[9] = "";        /* Current RTP DES key if currtpkey[0] !=
 static char curvatkey[9] = "";        /* Current VAT DES key if currtpkey[0] != 0 */
 static char curideakey[17] = "";      /* Current IDEA key if curideakey[0] != 0 */
 static char curpgpkey[17] = "";       /* Current PGP key if curpgpkey[0] != 0 */
-static char *curotp = NULL;	      /* Key file buffer */
-#ifdef TINY_PACKETS
-static int sound_packet;	      /* Ideal samples/packet */
+static char curblowfish_spec = FALSE; /* Nonzero if Blowfish key specified */
+#ifdef CRYPTO
+static BF_KEY curblowfishkey;	      /* Blowfish key */
 #endif
+static char *curotp = NULL;	      /* Key file buffer */
+static int sound_packet;	      /* Ideal samples/packet */
 static struct soundbuf *pgpsb = NULL; /* PGP key sound buffer, if any */
 static LONG pgpsbl;		      /* Length of PGP key sound buffer data */
 #ifdef HALF_DUPLEX
@@ -97,8 +113,79 @@ static long usticks;		      /* Microseconds per clock tick */
 extern int HPTermHookChar;
 #endif
 
+/* Audio input and control device file names. */
+
+#ifdef AUDIO_DEVICE_FILE
+extern char *devAudioInput, *devAudioControl;
+#endif
+
 #define ucase(x)    (islower(x) ? toupper(x) : (x))
 
+#ifndef USE_CURSES
+
+#ifndef UNIX420
+#define UNIX5
+#endif
+
+#ifdef UNIX420
+#include <sgtty.h>
+#endif /* UNIX420 */
+
+#ifdef UNIX5
+#include <termio.h>
+static struct termio old_term_params;
+#endif /* UNIX5 */
+
+/* Set raw mode on terminal file.  Basically, get the terminal into a
+   mode in which all characters can be read as they are entered.  CBREAK
+   mode is not sufficient.  */
+
+static void tty_rawmode()
+{
+#ifdef UNIX420
+   struct sgttyb arg;
+
+   ioctl(fileno(stdin), TIOCGETP, &arg);	  /* get basic parameters */
+   arg.sg_flags |= RAW; 	      /* set raw mode */
+   arg.sg_flags &= ~ECHO;	      /* clear echo mode */
+   ioctl(fileno(stdin), TIOCSETP, &arg);	  /* set it */
+#endif /* UNIX420 */
+
+#ifdef UNIX5
+   struct termio term_params;
+
+   ioctl(fileno(stdin), TCGETA, &old_term_params);
+   term_params = old_term_params;
+   term_params.c_iflag &= ~(ICRNL|IXON|IXOFF);	/* no cr translation */
+   term_params.c_iflag &= ~(ISTRIP);   /* no stripping of high order bit */
+   term_params.c_oflag &= ~(OPOST);    /* no output processing */	
+   term_params.c_lflag &= ~(ISIG|ICANON|ECHO); /* raw mode */
+   term_params.c_cc[4] = 1;  /* satisfy read after 1 char */
+   ioctl(fileno(stdin), TCSETAF, &term_params);
+#endif /* UNIX5 */
+/*printf("\n(raw)\n");*/
+}
+
+/* Restore tty mode */
+
+static void tty_normode()
+{
+#ifdef UNIX420
+struct sgttyb arg;
+
+   ioctl(fileno(stdin), TIOCGETP, &arg);	  /* get basic parameters */
+   arg.sg_flags &= ~RAW;	      /* clear raw mode */
+   arg.sg_flags |= ECHO;	      /* set echo mode */
+   ioctl(fileno(stdin), TIOCSETP, &arg);	  /* set it */
+#endif /* UNIX420 */
+
+#ifdef UNIX5
+   ioctl(fileno(stdin), TCSETAF, &old_term_params);
+#endif /* UNIX5 */
+/*printf("\n(cooked)\n");*/
+}
+#endif
+
 /*  ULARM  --  Wrapper for setitimer() that looks like alarm()
 	       but accepts a time in microseconds.  */
 
@@ -139,8 +226,69 @@ static void gsmcomp(sb)
     }
 
     /* Hide original uncompressed buffer length in first 2 bytes of buffer. */
+    
+#ifdef OBSOLETE
 
-    *((short *) sb->buffer.buffer_val) = sb->buffer.buffer_len;
+    /* Allow me to explain the following if statement.	Speak
+       Freely for Unix prior to release 6.1e contained a bug
+       which caused it to both send (here) and receive
+       (in speaker.c) GSM packets with the embedded length
+       in host byte order.  This went undetected for a long
+       time because all the Unix machines I have access to
+       for testing are big-endian and thus interoperated
+       correctly with one another and the Windows version
+       (which does handle the byte order correctly).  Many
+       Linux and FreeBSD users only speak to one another,
+       and thus never noticed the problem.  It was only when
+       a little-endian Unix machine attempted to use GSM
+       compression in a connection with a Windows or
+       big-endian Unix machine that what came to be called
+       the "deafening static" problem manifested itself.
+       The fix is obvious, but we have to get there in
+       a cautious manner to avoid breaking interoperability
+       during the transition.  So here's the plan: starting
+       with Unix release 6.1e and the corresponding Windows
+       release, the GSM receiving code in both versions
+       will accept the GSM packet length field in
+       *either* byte order (a byte-swapped value in this
+       field can be detected unambiguously because if
+       byte-swapped it will exceed the actual length of
+       the packet, which is otherwise impossible).  Users
+       who install these releases will be able to receive
+       GSM sound from any version of Speak Freely on any
+       platform, present or future.  Version 6.1e for
+       Unix will, however, continue to send the GSM length
+       field in host byte order.  Why?	Because otherwise
+       Linux and FreeBSD users who communicate with one
+       another would both have to install 6.1e before
+       they could communicate once again.  To make the
+       transition smoother, 6.1e offers a "-tc" option,
+       which causes the GSM length field to be sent
+       in the correct byte order.  This option should be
+       set whenever a Unix user wishes to communicate
+       with a Windows user in GSM mode.  It will also
+       fix connections to big-endian Unix versions 
+       prior to 6.1e.  Once a majority of the installed
+       base is on the versions which don't care about
+       byte order in this field, sending in network byte
+       order will be the default.  */
+
+    if (gsm_byte_order_kludge) {
+#endif
+    *((short *) sb->buffer.buffer_val) = htons((short) sb->buffer.buffer_len);
+#ifdef OBSOLETE
+    } else {
+	*((short *) sb->buffer.buffer_val) = sb->buffer.buffer_len;
+    }
+#endif
+#ifdef BYTE_SWAP_DEBUG
+    if (gsm_byte_order_debug) {
+	unsigned char *bsp = (unsigned char *) sb->buffer.buffer_val;
+
+	bsp[0] = sb->buffer.buffer_len & 0xFF;
+	bsp[1] = (sb->buffer.buffer_len >> 8) & 0xFF;
+    }
+#endif
     sb->buffer.buffer_len = l + sizeof(short);
 }
 
@@ -229,6 +377,7 @@ static int sendrtpctrl(d, destaddr, msg, msgl, rtppkt)
 {
     char aux[1024];
 
+#ifdef CRYPTO
     if ((rtp || vat) && d->rtpdeskey[0]) {
 	int vlen;
 	des_cblock key;
@@ -270,6 +419,7 @@ static int sendrtpctrl(d, destaddr, msg, msgl, rtppkt)
 	    (des_cblock *) ivec, DES_ENCRYPT);
 	msg = aux;
     }
+#endif
     return sendto(sock, msg, msgl, 0, destaddr, sizeof d->ctrl);
 }
 
@@ -337,6 +487,10 @@ static int addest(host)
     bcopy(curvatkey, d->vatdeskey, 9);
     bcopy(curideakey, d->ideakey, 17);
     bcopy(curpgpkey, d->pgpkey, 17);
+    d->blowfish_spec = curblowfish_spec;
+#ifdef CRYPTO
+    bcopy(&curblowfishkey, &(d->blowfishkey), sizeof(BF_KEY));
+#endif
     d->otp = curotp;
     d->time_to_live = ttl;
     d->name.sin_family = AF_INET;
@@ -378,6 +532,7 @@ static int addest(host)
 	sdes[0] = pid;
     }
 
+#ifdef CRYPTO
     if (pgpsb != NULL) {
 	int i;
 
@@ -396,6 +551,7 @@ static int addest(host)
 /* No need to sleep once ack from far end confirms key decoded. */
 sleep(7);
     }
+#endif
     return TRUE;
 }
 
@@ -497,7 +653,10 @@ static int sendpkt(sb)
 	    }
 	}
 #endif
-	if (d->deskey[0] || d->ideakey[0] || d->pgpkey[0] || d->otp != NULL) {
+
+#ifdef CRYPTO
+	if (d->deskey[0] || d->ideakey[0] || d->blowfish_spec ||
+	    d->pgpkey[0] || d->otp != NULL) {
 	    soundbuf ebuf;
 	    int i;
 	    LONG slen, plen = 0;
@@ -597,6 +756,29 @@ static int sendpkt(sb)
 		ebuf.compression |= fEncIDEA;
 	    }
 
+	    /* Blowfish encryption. */
+
+	    if (d->blowfish_spec) {
+		unsigned char iv[8];
+
+		bzero(iv, sizeof(iv));
+
+                /* If we're Blowfish encrypting we must round the size of
+		   the data to be sent to be a multiple of 8 so that
+		   the entire Blowfish frame is sent. */
+
+		slen = (slen + 7) & (~7);
+		pktlen = slen + (sizeof(struct soundbuf) - BUFL);
+		if (debugging) {
+                    fprintf(stderr, "Encrypting %d bytes with Blowfish key.\r\n",
+			    slen);
+		}
+		BF_cbc_encrypt((unsigned char *) ebuf.buffer.buffer_val,
+			       (unsigned char *) ebuf.buffer.buffer_val,
+			       slen, &(d->blowfishkey), iv, BF_ENCRYPT);
+		ebuf.compression |= fEncBF;
+	    }
+
 	    /* PGP encryption. */
 
 	    if (d->pgpkey[0]) {
@@ -645,14 +827,16 @@ static int sendpkt(sb)
 		ebuf.compression = htonl(ebuf.compression);
 		ebuf.buffer.buffer_len = htonl(ebuf.buffer.buffer_len);
 	    }
-	    for (i = 0; i < (lpc10comp ? robust : 1); i++) {
+	    for (i = 0; i < (lpc10compress ? robust : 1); i++) {
 		if (sendto(sock, (char *) &ebuf, pktlen,
 		    0, (struct sockaddr *) &(d->name), sizeof d->name) < 0) {
-                    perror("sending datagram message");
+                    perror("sending encrypted audio");
 		    return FALSE;
 		}
 	    }
-	} else {
+	} else
+#endif
+	{
 	    int i, wasf = FALSE;
 
 	    if (!vat && !rtp) {
@@ -669,10 +853,10 @@ static int sendpkt(sb)
 		sb->buffer.buffer_len = htonl(sb->buffer.buffer_len);
 	    }
 
-	    for (i = 0; i < (lpc10comp ? robust : 1); i++) {
+	    for (i = 0; i < (lpc10compress ? robust : 1); i++) {
 		if (sendto(sock, (char *) sb, pktlen,
 		    0, (struct sockaddr *) &(d->name), sizeof d->name) < 0) {
-                    perror("sending datagram message");
+                    perror("sending audio");
 		    return FALSE;
 		}
 	    }
@@ -747,8 +931,8 @@ static int getaudio()
 #endif
 
     if (i <= 0) {
-            fprintf(stderr, "%s: unable to initialise audio.\n", progname);
-	    return FALSE;
+        fprintf(stderr, "%s: unable to initialise audio.\n", progname);
+	return FALSE;
     }
     havesound = TRUE;
     if (agc) {
@@ -804,13 +988,24 @@ static void exiting()
 {
     struct destination *d;
 
+    freeaudio();
+
+    if (audioDumpFile != NULL) {
+	fclose(audioDumpFile);
+    }
+
 #ifdef PUSH_TO_TALK
     if (rawmode) {
         fprintf(stderr, "\r      \r");
-	fcntl(fileno(stdin), F_SETFL, 0);
+#ifdef USE_CURSES
 	nocbreak();
 	echo();
 	endwin();
+	fcntl(fileno(stdin), F_SETFL, 0);
+#else
+	tty_normode();
+	fcntl(fileno(stdin), F_SETFL, 0);
+#endif
     }
 #endif
 
@@ -850,10 +1045,22 @@ static void exiting()
 	}
     }
 
-    freeaudio();
     exit(0);
 }
 
+/*  CHANGEMODE	--  Reset state when talk/pause/quiet mode changes.  */
+
+static void changemode()
+{
+    if (talking) {
+	spurt = TRUE;
+	/* Discard all backlog sound input. */
+	soundflush();
+    } else {
+	sqwait = 0;		      /* Reset squelch delay */
+    }
+}
+
 /*  TERMCHAR  --  Check for special characters from console when
 		  in raw mode.	*/
 
@@ -867,6 +1074,73 @@ static void termchar(ch)
     }
 }
 
+/*  CHATCHAR  --  Test for text chat request character.  */
+
+static int chatchar(ch)
+  int ch;
+{
+    if ((!rtp && !vat) && ch == '.') {
+	char s[133];
+
+	if (rawmode) {
+#ifdef USE_CURSES
+	    nocbreak();
+	    echo();
+	    endwin();
+#else
+	    tty_normode();
+#endif
+	}
+	fcntl(fileno(stdin), F_SETFL, 0);
+	if (push) {
+	    talking = 0;
+	}
+
+        fprintf(stderr, "\rChat:  ");
+	while (TRUE) {
+	    if (fgets(s, (sizeof s) - 1, stdin) != NULL) {
+		struct destination *d;
+
+		while (strlen(s) > 0 && isspace(s[strlen(s) - 1])) {
+		    s[strlen(s) - 1] = 0;
+		}
+		for (d = dests; d != NULL; d = d->dnext) {
+		    char v[256];
+		    int l;
+
+		    l = rtp_make_app(v, ssrc, TRUE, RTCP_APP_TEXT_CHAT, s);
+		    /* Set Speak Freely protocol flag in packet */
+		    v[0] = (v[0] & 0x3F) | (1 << 6);
+		    if (sendrtpctrl(d, (struct sockaddr *) &(d->ctrl),
+				 v, l, TRUE) < 0) {
+                        perror("sending text chat packet");
+		    }
+		}
+/*
+fprintf(stderr, "\nOut of chat mode, input(%d) = \"%s\"\n", strlen(s), s);
+*/
+		fprintf(stderr, talking ? (squelch > 0 ?
+                        "\rQuiet: " : "\rTalk:   ") : "\rPause: ");
+
+		if (rawmode) {
+#ifdef USE_CURSES
+		    initscr();
+		    noecho();
+		    cbreak();
+#else
+		    tty_rawmode();
+#endif
+		}
+		changemode();
+		break;
+	    }
+	}
+	fcntl(fileno(stdin), F_SETFL, talking ? O_NDELAY : 0);
+	return TRUE;
+    }
+    return FALSE;
+}
+
 /*  SENDFILE  --  Send a file or, if the file name is NULL or a
 		  single period, send real-time sound input. */
 
@@ -888,8 +1162,6 @@ static int sendfile(f)
 	ularm(TimerStep);
     }
 
-#ifdef TINY_PACKETS
-
     /* Compute the number of sound samples needed to fill a
        packet of TINY_PACKETS bytes. */
 
@@ -916,7 +1188,6 @@ static int sendfile(f)
 
 #ifdef SHOW_PACKET_SIZE
     printf("Samples per packet = %d\n", sound_packet);
-#endif
 #endif
     lread = sound_packet;
 
@@ -974,11 +1245,7 @@ static int sendfile(f)
 
 	    (bread = nread =
 		fread(buf + (toasted ? 2 : 0), 1,
-#ifdef TINY_PACKETS
 				  lread,
-#else
-				  (gsmcompress ? 1600 : BUFL),
-#endif
 				  afile)) > 0) {
 	    netbuf.compression = fProtocol | (ring ? (fSetDest | fDestSpkr) : 0);
 	    ring = FALSE;
@@ -990,19 +1257,10 @@ static int sendfile(f)
 		netbuf.compression |= fCompGSM;
 	    } else {
 		if (compressing) {
-#ifdef OLD2X
-		    int i;
-
-		    nread /= 2;
-		    for (i = 1; i < nread; i++) {
-			buf[i] = buf[i * 2];
-		    }
-#else
 		    int is = nread, os = nread / 2;
 
 		    rate_flow(buf, buf, &is, &os);
 		    nread = os;
-#endif
 		    netbuf.compression |= fComp2X;
 		}
 		netbuf.buffer.buffer_len = nread;
@@ -1041,11 +1299,7 @@ static int sendfile(f)
 	       to neither convention, we just guess based on the number
 	       of samples in the packet and a fudge factor. */
 
-#ifdef TINY_PACKETS
 #define kOverhead   8000
-#else
-#define kOverhead   3000
-#endif
 
 #ifdef BSD_like
 	    ftime(&t2);
@@ -1066,7 +1320,11 @@ printf("Packet time %d Delta t = %d Error = %d, Corr = %d, Wait = %d\n",
 	atime, dt, error, corr, et);
 #endif
 	    } else {
+#ifdef __FreeBSD__
+		corr = 80000;
+#else
 		corr = 8000;
+#endif
 		et = ((bread * 125) -
 		      1000 * (((t2.time - t1.time) * 1000) +
 			       (t2.millitm - t1.millitm))) - kOverhead;
@@ -1105,7 +1363,7 @@ printf("Packet time %d Delta t = %d Error = %d, Corr = %d, Wait = %d\n",
 	    }
 	    if (debugging && !vat && !rtp) {
                 fprintf(stderr, "Sent %d samples from %s in %d bytes.\r\n",
-			bread, f, netbuf.buffer.buffer_len);
+			bread, f, ntohl(netbuf.buffer.buffer_len));
 	    }
 	}
 
@@ -1124,22 +1382,30 @@ printf("Packet time %d Delta t = %d Error = %d, Corr = %d, Wait = %d\n",
 	    int l;
 
 	    fprintf(stderr,
-                "Space bar switches talk/pause, Esc or \"q\" to quit\nPause: ");
+                "Space bar switches talk/pause, Esc or \"q\" to quit, \".\" for chat text\nPause: ");
 	    fflush(stderr);
+#ifdef USE_CURSES
 	    initscr();
 	    noecho();
 	    cbreak();
+#else
+	    tty_rawmode();
+#endif
 	    rawmode = TRUE;
 	    while (TRUE) {
-		if (l = read(fileno(stdin), &c, 1) == 1 || errno != EINTR) {
+		while (TRUE) {
+		    if (l = read(fileno(stdin), &c, 1) == 1 || errno != EINTR) {
+			break;
+		    }
+		}   
+		if (l == 1 && !chatchar(c)) {
 		    break;
 		}
-	    }	
+	    }
 	    if (l != 1) {
                 perror("waiting for first Pause/Talk character");
 	    }
 	    termchar(c);
-/*          fprintf(stderr, "\rTalk:  "); */
             fprintf(stderr, squelch > 0 ? "\rQuiet:  " : "\rTalk:   ");
 	    fcntl(fileno(stdin), F_SETFL, O_NDELAY);
 	    talking = TRUE;
@@ -1150,31 +1416,34 @@ printf("Packet time %d Delta t = %d Error = %d, Corr = %d, Wait = %d\n",
 	/* Send real-time sound. */
 
 #ifdef Solaris
-	setaubufsize(
-#ifdef TINY_PACKETS
-	    sound_packet
-#else
-	    sizeof buf
-#endif
-		    );
+	setaubufsize(sound_packet);
 #endif
 
 	if (!getaudio()) {
             fprintf(stderr, "Unable to initialise audio.\n");
 	    return 2;
 	}
+	changemode();
 	while (TRUE) {
 	    int soundel = 0;
 	    unsigned char *bs = (unsigned char *) buf;
 
 	    if (havesound) {
-		soundel = soundgrab(buf,
-#ifdef TINY_PACKETS
-					 sound_packet
-#else
-					 sizeof buf
-#endif
-				   );
+
+                /* Obtain a packet's worth of real-time audio from
+		   the audio input hardware. */
+
+		soundel = soundgrab(buf, sound_packet);
+
+		/* If a raw audio dump file (-w option) is specified,
+		   dump whatever we got from the audio input to the
+		   designated file.  This can come in handy when
+                   debugging audio drivers which don't provide samples
+		   in the expected 8 bit mu-law mode. */
+
+		if ((audioDumpFile != NULL) && (soundel > 0)) {
+		    fwrite(buf, 1, soundel, audioDumpFile);
+		}
 	    }
 
 #ifdef PUSH_TO_TALK
@@ -1190,6 +1459,9 @@ printf("Packet time %d Delta t = %d Error = %d, Corr = %d, Wait = %d\n",
 #ifdef HEWLETT_PACKARD
 		    c = HPTermHookChar;
 #endif
+		    if (rlen > 0 && chatchar(c)) {
+			continue;
+		    }
 		    termchar(c);
 		    talking = !talking;
 		    fflush(stderr);
@@ -1211,20 +1483,19 @@ printf("Packet time %d Delta t = %d Error = %d, Corr = %d, Wait = %d\n",
 		    fprintf(stderr, talking ? (squelch > 0 ?
                             "\rQuiet: " : "\rTalk:   ") : "\rPause: ");
 		    fcntl(fileno(stdin), F_SETFL, talking ? O_NDELAY : 0);
+#ifdef OBSOLETE
 		    if (talking) {
 			spurt = TRUE;
 			/* Discard all backlog sound input. */
 			soundflush();
+fprintf(stderr, "\n**flush**\n");
 		    } else {
 			sqwait = 0;   /* Reset squelch delay */
 		    }
+#endif
+		    changemode();
 		    break;
 		} else {
-/*
-if (rlen == -1) {
-perror("Read poll"); fflush(stderr);
-}
-*/
 		    if (rlen == -1 && errno == EINTR) {
 			continue;
 		    }
@@ -1317,19 +1588,10 @@ perror("Read poll"); fflush(stderr);
 
 		    ring = FALSE;
 		    if (compressing) {
-#ifdef OLD2X
-			int i;
-
-			soundel /= 2;
-			for (i = 1; i < soundel; i++) {
-			    buf[i] = buf[i * 2];
-			}
-#else
 			int is = soundel, os = soundel / 2;
 
 			rate_flow(buf, buf, &is, &os);
 			soundel = os;
-#endif
 			netbuf.compression |= fComp2X;
 		    }
 		    netbuf.buffer.buffer_len = soundel;
@@ -1397,10 +1659,15 @@ static void makeSessionKey(key, seed, mode)
     MD5Init(&md5c);
     MD5Update(&md5c, s, strlen(s));
     MD5Final(md5key1, &md5c);
+#ifdef CRYPTO
     init_idearand(md5key, md5key1, time(NULL));
+#define nextrand    idearand()
+#else
+#define nextrand    (md5key[j] ^ md5key1[j])
+#endif
     if (mode) {
 	for (j = k = 0; j < 16; j++) {
-	    unsigned char rb = idearand();
+	    unsigned char rb = nextrand;
 
 #define Rad16(x) ((x) + 'A')
 	    key[k++] = Rad16((rb >> 4) & 0xF);
@@ -1413,10 +1680,13 @@ static void makeSessionKey(key, seed, mode)
 	key[--k] = 0;
     } else {
 	for (j = 0; j < 16; j++) {
-	    key[j] = idearand();
+	    key[j] = nextrand;
 	}
     }
+#ifdef CRYPTO
     close_idearand();
+#endif
+#undef nextrand
 }
 
 /*  PROG_NAME  --  Extract program name from argv[0].  */
@@ -1441,33 +1711,64 @@ static void usage()
 #ifdef PUSH_TO_TALK
     V fprintf(stderr, "           -A         Always transmit unless squelched\n");
     V fprintf(stderr, "     *     -B         Push to talk using keyboard\n");
+#endif
+#ifdef CRYPTO
+    V fprintf(stderr, "           -BFkey     Blowfish encrypt with key\n");
 #endif
     V fprintf(stderr, "           -C         Compress subsequent sound\n");
     V fprintf(stderr, "           -D         Enable debug output\n");
+#ifdef CRYPTO
     V fprintf(stderr, "           -E[key]    Emit session key string\n");
+#endif
     V fprintf(stderr, "           -F         ADPCM compression\n");
     V fprintf(stderr, "           -G         Automatic gain control\n");
+#ifdef CRYPTO
     V fprintf(stderr, "           -Ikey      IDEA encrypt with key\n");
     V fprintf(stderr, "           -Kkey      DES encrypt with key\n");
+#endif
     V fprintf(stderr, "           -L         Remote loopback\n");
     V fprintf(stderr, "           -LPC       LPC compression\n");
     V fprintf(stderr, "           -LPC10Rn   LPC-10 compression, n copies\n");
     V fprintf(stderr, "     *     -M         Manual record gain control\n");
     V fprintf(stderr, "           -N         Do not compress subsequent sound\n");
+#ifdef CRYPTO
     V fprintf(stderr, "           -Ofile     Use file as key file\n");
+#endif
     V fprintf(stderr, "           -Phostname[:port] Party line, add host to list\n");
     V fprintf(stderr, "     *     -Q         Disable debug output\n");
     V fprintf(stderr, "           -R         Ring--force volume, output to speaker\n");
     V fprintf(stderr, "           -RTP       Use Internet Real-Time Protocol\n");
     V fprintf(stderr, "           -Sn,t      Squelch at level n (0-32767), timeout t milliseconds\n");
+ 
     V fprintf(stderr, "     *     -T         Telephone (GSM) compression\n");
+#ifdef OBSOLETE
+    V fprintf(stderr, "           -TC        GSM with corrected x86 byte order\n");
+#endif
+#ifdef BYTE_SWAP_DEBUG
+    V fprintf(stderr, "           -TD        Debug: send GSM with swapped byte order\n");
+#endif
     V fprintf(stderr, "           -U         Print this message\n");
     V fprintf(stderr, "           -VAT       Use VAT protocol\n");
+    V fprintf(stderr, "           -Wfile     Write raw audio input to file\n");
+#ifdef AUDIO_DEVICE_FILE
+    V fprintf(stderr, "           -Yindev[:ctldev] Override default audio device file name or specify open #fd\n");
+#endif
+#ifdef CRYPTO
     V fprintf(stderr, "           -Z\"user..\" Send PGP session key for user(s)\n");
+#endif
     V fprintf(stderr, "\n");
     V fprintf(stderr, "by John Walker\n");
-    V fprintf(stderr, "   E-mail: kelvin@fourmilab.ch\n");
-    V fprintf(stderr, "   WWW:    http://www.fourmilab.ch/\n");
+    V fprintf(stderr, "   http://www.fourmilab.ch/\n");
+#ifndef CRYPTO
+    V fprintf(stderr, "\n");
+    V fprintf(stderr, "Note: This version of Speak Freely was built with\n");
+    V fprintf(stderr, "      encryption code removed to permit redistribution\n");
+    V fprintf(stderr, "      without concern for export control and regulations\n");
+    V fprintf(stderr, "      concerning encryption technology in some jurisdictions.\n");
+    V fprintf(stderr, "      You can download a version of Speak Freely with\n");
+    V fprintf(stderr, "      full encryption from:\n");
+    V fprintf(stderr, "             http://www.fourmilab.ch/speakfree/unix/\n");
+#endif
 }
 
 /*  Main program.  */
@@ -1484,10 +1785,14 @@ main(argc, argv)
     char s[1024];
 
     progname = prog_name(argv[0]);
-    gethostname(hostname, sizeof hostname);
-    if (strlen(hostname) > 15) {
-	hostname[15] = 0;
+    if (gethostname(hostname, sizeof hostname) == -1) {
+        strcpy(hostname, "localhost");
+    } else {
+	if (strlen(hostname) > 15) {
+	    hostname[15] = 0;
+	}
     }
+
 #ifdef LINUX_FPU_FIX
     __setfpucw(_FPU_IEEE);	      /* Mask floating point interrupts to
 					 enable standard IEEE processing.
@@ -1495,9 +1800,7 @@ main(argc, argv)
 					 default, so this is needed only on
 					 older libraries. */
 #endif
-#ifndef OLD2X
     rate_start(8000, 4000);
-#endif
     gsmh = gsm_create();
 
     /* Initialise LPC decoding. */
@@ -1515,15 +1818,15 @@ main(argc, argv)
 
 #ifdef HALF_DUPLEX
     {
-	struct hostent *h;
-	char host[512];
-
-	gethostname(host, sizeof host);
-	h = gethostbyname(host);
-	bcopy((char *) (h->h_addr), (char *) &localhost,
-	    sizeof localhost);
+#ifndef INADDR_LOOPBACK
+	/* The standard local host loopback address is supposed to be
+	   defined in <netinet/in.h>, but just in case some screwball
+           system doesn't, let's include our own definition here. */
+#define INADDR_LOOPBACK  0x7F000001
+#endif
+	localhost.s_addr = htonl(INADDR_LOOPBACK);
 #ifdef HDX_DEBUG
-        fprintf(stderr, "%s: local host %s = %s\n", progname, host,
+        fprintf(stderr, "%s: local host %s\n", progname,
 		inet_ntoa(localhost));
 #endif
     }
@@ -1586,9 +1889,35 @@ main(argc, argv)
 		    break;
 
                 case 'B':             /* -B  -- Push to talk (button) */
-		    if (isatty(fileno(stdin))) {
-			push = TRUE;
+#ifdef CRYPTO
+				      /* -BFkey  -- Set Blowfish key */ 
+                    if (ucase(op[1]) == 'F') {
+			if (strlen(op + 2) == 0) {
+			    curblowfish_spec = FALSE;
+			} else {
+			    unsigned char bfvec[16];
+
+			    curblowfish_spec = TRUE;
+			    MD5Init(&md5c);
+			    MD5Update(&md5c, op + 2, strlen(op + 2));
+			    MD5Final(bfvec, &md5c);
+			    BF_set_key(&(curblowfishkey), 16, bfvec);
+			    if (debugging) {
+                                fprintf(stderr, "Blowfish key:");
+				for (j = 0; j < 16; j++) {
+                                    fprintf(stderr, " %02X", (bfvec[j] & 0xFF));
+				}
+                                fprintf(stderr, "\n");
+			    }
+			}
+		    } else {
+#endif
+			if (isatty(fileno(stdin))) {
+			    push = TRUE;
+			}
+#ifdef CRYPTO
 		    }
+#endif
 		    break;
 #endif
 
@@ -1600,10 +1929,12 @@ main(argc, argv)
 		    debugging = TRUE;
 		    break;
 
+#ifdef CRYPTO
                 case 'E':             /* -E  --  Emit session key and exit */
 		    makeSessionKey(s, op + 1, TRUE);
                     printf("%s\n", s);
 		    return 0;
+#endif
 
                 case 'F':             /* -F -- ADPCM compression */
 		    adpcmcompress = TRUE;
@@ -1614,6 +1945,7 @@ main(argc, argv)
 		    agc = TRUE;
 		    break;
 
+#ifdef CRYPTO
                 case 'I':             /* -Ikey  --  Set IDEA key */
 		    if (strlen(op + 1) == 0) {
 			curideakey[0] = FALSE;
@@ -1678,6 +2010,7 @@ main(argc, argv)
 			}
 		    }
 		    break;
+#endif
 
                 case 'L':             /* -L      --  Remote loopback */
 				      /* -LPC	 --  LPC compress sound */
@@ -1719,6 +2052,7 @@ main(argc, argv)
 		    adpcmcompress = FALSE;
 		    break;
 
+#ifdef CRYPTO
                 case 'O':             /* -Ofile -- Use file as key file */
 		    if (op[1] == 0) {
 			curotp = NULL; /* Switch off key file */
@@ -1754,6 +2088,7 @@ main(argc, argv)
 			}
 		    }
 		    break;
+#endif
 
                 case 'P':             /* -Phost  --  Copy output to host  */
 		    if (!addest(op + 1)) {
@@ -1796,6 +2131,23 @@ main(argc, argv)
                 case 'T':             /* -T -- Telephone (GSM) compression */
 		    gsmcompress = TRUE;
 		    lpccompress = lpc10compress = adpcmcompress = FALSE;
+#ifdef OBSOLETE
+                    if (ucase(op[1]) == 'C') {
+			gsm_byte_order_kludge = TRUE;
+		    }
+#endif
+#ifdef BYTE_SWAP_DEBUG
+		    /* This can be used if you absolutely have to communicate
+		       with somebody running a version prior to 6.1e on little-endian
+		       hardware.  This brakes the byte order of GSM packets to
+		       compensate for a bug in earlier versions.  Note that you
+		       should only set this for compatibility with old Unix
+		       versions--the problem this works around was never in
+		       the Windows version. */
+                    if (ucase(op[1]) == 'D') {
+			gsm_byte_order_debug = TRUE;
+		    }
+#endif
 		    break;
 
                 case 'U':             /* -U  --  Print usage information */
@@ -1810,6 +2162,25 @@ main(argc, argv)
 		    }
 		    break;
 
+                case 'W':             /* -Wfile -- Write raw audio to file */
+                    audioDumpFile = fopen(op + 1, "w");
+		    if (audioDumpFile == NULL) {
+                        perror("Cannot create raw audio dump (-w option) file");
+			return 2;
+		    }
+		    break;
+
+#ifdef AUDIO_DEVICE_FILE
+                case 'Y':             /* -Yindev:[ctldev] -- Specify audio input and control device file names
+							     or #open_fd. */
+		    devAudioInput = op + 1;
+                    if (strchr(op + 1, ':') != NULL) {
+                        devAudioControl = strchr(op + 1, ':') + 1;
+		    }
+		    break;
+#endif
+
+#ifdef CRYPTO
                 case 'Z':             /* -Z"user1 user2..."  --  Send PGP
 						 encrypted session key to
 						 named users */
@@ -1866,12 +2237,10 @@ main(argc, argv)
 #ifdef PGP_DEBUG
                             fprintf(stderr, "PGP key buffer length: %d\n", flen);
 #endif
-#ifdef TINY_PACKETS
 			    if (flen > (TINY_PACKETS -
 				    (sizeof(soundbuf) - BUFL))) {
                                 fprintf(stderr, "Warning: PGP key message exceeds %d packet size.\n", TINY_PACKETS);
 			    }
-#endif
 			    if (pgpsb != NULL) {
 				free(pgpsb);
 			    }
@@ -1900,6 +2269,7 @@ main(argc, argv)
 			unlink(f);
 		    }
 		    break;
+#endif
 	    }
 	} else {
 
@@ -1911,13 +2281,10 @@ main(argc, argv)
 	    }
 
 	    if (rtp || vat) {
-		if (curideakey[0] || curpgpkey[0] || curotp != NULL) {
-                    fprintf(stderr, "IDEA, PGP, and Key File encryption cannot be used\n    with -RTP or -VAT protocols.\n");
-		    return 2;
-		}
-#ifdef ZZZZ
-		if (vat && curkey[0]) {
-                    fprintf(stderr, "DES encryption not implemented for VAT protocol.\n");
+#ifdef CRYPTO
+		if (curideakey[0] || curpgpkey[0] ||
+		    curblowfish_spec || curotp != NULL) {
+                    fprintf(stderr, "Blowfish, IDEA, PGP, and Key File encryption cannot be used\n    with -RTP or -VAT protocols.\n");
 		    return 2;
 		}
 #endif
@@ -1958,6 +2325,8 @@ main(argc, argv)
     exiting();
     gsm_destroy(gsmh);
     lpc_end();
+#ifdef CRYPTO
     desdone();
+#endif
     return 0;
 }
